# coding: utf-8

# docs
from __future__ import annotations

import os
import random
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union
from uuid import UUID

from tqdm import tqdm

import supervisely.imaging.image as sly_image
from supervisely._utils import batched
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.pointcloud.pointcloud_api import PointcloudInfo
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.io.fs import dir_exists, get_file_name, list_files, mkdir, touch
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.pointcloud_annotation.constants import OBJECT_KEY
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import (
    PointcloudEpisodeAnnotation,
)
from supervisely.project.pointcloud_project import PointcloudDataset, PointcloudProject
from supervisely.project.project import OpenMode
from supervisely.project.project import read_single_project as read_project_wrapper
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_type import ProjectType
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress, tqdm_sly
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap

KITTI_ITEM_DIR_NAME = "velodyne"


class EpisodeItemPaths(NamedTuple):
    #: :class:`str`: Full pointcloud file path of item
    pointcloud_path: str

    #: :class:`str`: Path to related images directory of item
    related_images_dir: str

    #: :class:`int`: Index of frame in episode annotation of dataset
    frame_index: int


class EpisodeItemInfo(NamedTuple):
    #: :class:`str`: Item's dataset name
    dataset_name: str

    #: :class:`str`: Item name
    name: str

    #: :class:`str`: Full pointcloud file path of item
    pointcloud_path: str

    #: :class:`str`: Path to related images directory of item
    related_images_dir: str

    #: :class:`int`: Index of frame in episode annotation of dataset
    frame_index: int


class PointcloudEpisodeDataset(PointcloudDataset):
    #: :class:`str`: Items data directory name
    item_dir_name = "pointcloud"

    #: :class:`str`: Annotations directory name
    ann_dir_name = "ann"

    #: :class:`str`: Items info directory name
    item_info_dir_name = "pointcloud_info"

    #: :class:`str`: Related images directory name
    related_images_dir_name = "related_images"

    #: :class:`str`: Segmentation masks directory name
    seg_dir_name = None

    item_info_type = PointcloudInfo
    annotation_class = PointcloudEpisodeAnnotation

    @property
    def ann_dir(self) -> None:
        raise NotImplementedError(
            f"{type(self).__name__} object don't have correct path for 'ann_dir' property. \
            Use 'get_ann_path()' method instead of this."
        )

    def get_item_paths(self, item_name: str) -> EpisodeItemPaths:
        return EpisodeItemPaths(
            pointcloud_path=self.get_pointcloud_path(item_name),
            related_images_dir=self.get_related_images_path(item_name),
            frame_index=self.get_frame_idx(item_name),
        )

    def get_ann_path(self) -> str:
        return os.path.join(self.directory, "annotation.json")

    def get_ann(
        self, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudEpisodeAnnotation:
        """
        Read pointcloud annotation of item from json.

        :param item_name: Pointcloud name.
        :type item_name: str
        :param project_meta: Project Meta.
        :type project_meta: :class:`ProjectMeta<supervisely.ProjectMeta>`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`KeyIdMap<supervisely.video_annotation.key_id_map.KeyIdMap>`, optional
        :return: PointcloudEpisodeAnnotation object
        :rtype: :class:`PointcloudEpisodeAnnotation<supervisely.PointcloudEpisodeAnnotation>`
        :raises: :class:`RuntimeError` if item not found in the project
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_path = "/home/admin/work/supervisely/projects/pointcloud_project"
            project = sly.PointcloudProject(project_path, sly.OpenMode.READ)

            ds = project.datasets.get('ds1')

            annotation = ds.get_ann("PTC_0056")
            # Output: RuntimeError: Item PTC_0056 not found in the project.

            annotation = ds.get_ann("PTC_0056.pcd")
            print(type(annotation).__name__)
            # Output: PointcloudEpisodeAnnotation
        """
        ann_path = self.get_ann_path()
        return self.annotation_class.load_json_file(ann_path, project_meta, key_id_map)

    def get_ann_frame(self, item_name: str, annotation: PointcloudEpisodeAnnotation) -> Frame:
        frame_idx = self.get_frame_idx(item_name)
        if frame_idx is None:
            raise ValueError(f"Frame wasn't assigned to pointcloud with name {item_name}.")
        return annotation.frames.get(frame_idx)

    def get_frame_pointcloud_map_path(self) -> str:
        return os.path.join(self.directory, "frame_pointcloud_map.json")

    def set_ann(self, ann: PointcloudEpisodeAnnotation) -> None:
        if type(ann) is not self.annotation_class:
            raise TypeError(
                f"Type of 'ann' should be {self.annotation_class.__name__}, not a {type(ann).__name__}"
            )
        dst_ann_path = self.get_ann_path()
        dump_json_file(ann.to_json(), dst_ann_path)

    def _create(self):
        mkdir(self.item_dir)

    def _read(self):
        if not dir_exists(self.item_dir):
            message = f"Cannot read dataset '{self.name}': '{self.item_dir}' directory not found"
            if dir_exists(os.path.join(self.directory, KITTI_ITEM_DIR_NAME)):
                message = f"Cannot read dataset '{self.name}'. The item directory '{KITTI_ITEM_DIR_NAME}' was found. This appears to be a KITTI dataset and will be skipped."
            raise NotADirectoryError(message)

        try:
            item_paths = sorted(list_files(self.item_dir, filter_fn=self._has_valid_ext))
            item_names = sorted([os.path.basename(path) for path in item_paths])

            map_file_path = self.get_frame_pointcloud_map_path()
            if os.path.isfile(map_file_path):
                self._frame_to_pc_map = load_json_file(map_file_path)
            else:
                self._frame_to_pc_map = {
                    frame_index: item_names[frame_index] for frame_index in range(len(item_names))
                }

            self._pc_to_frame = {v: k for k, v in self._frame_to_pc_map.items()}
            self._item_to_ann = {name: self._pc_to_frame[name] for name in item_names}
        except Exception as ex:
            raise Exception(f"Cannot read dataset ({self.name}): {repr(ex)}")

    def add_item_file(
        self,
        item_name: str,
        item_path: str,
        frame: Optional[Union[str, int]] = None,
        _validate_item: Optional[bool] = True,
        _use_hardlink: Optional[bool] = False,
        item_info: Optional[NamedTuple] = None,
    ) -> None:
        """
        Adds given item file to dataset items directory, and adds given annotation to dataset ann directory. if ann is None, creates empty annotation file.

        :param item_name: Item name.
        :type item_name: str
        :param item_path: Path to the item.
        :type item_path: str
        :param frame: Frame number.
        :type frame: str or int, optional
        :param _validate_item: Checks input files format.
        :type _validate_item: bool, optional
        :param _use_hardlink: If True creates a hardlink pointing to src named dst, otherwise don't.
        :type _use_hardlink: bool, optional
        :param item_info: NamedTuple ImageInfo containing information about pointcloud.
        :type item_info: NamedTuple, optional
        :return: None
        :rtype: :class:`NoneType`
        :raises: :class:`Exception` if item_name already exists in dataset or item name has unsupported extension.
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            dataset_path = "/home/admin/work/supervisely/projects/episodes_project/episode_0"
            ds = sly.PointcloudEpisodeDataset(dataset_path, sly.OpenMode.READ)

            ds.add_item_file("PTC_777.pcd", "/home/admin/work/supervisely/projects/episodes_project/episode_0/pointcloud/PTC_777.pcd", frame=3)
        """
        if item_path is None and item_info is None:
            raise RuntimeError("No item_path or ann or item_info provided.")

        self._add_item_file(
            item_name,
            item_path,
            _validate_item=_validate_item,
            _use_hardlink=_use_hardlink,
        )
        self._add_ann_by_type(item_name, frame)
        self._add_item_info(item_name, item_info)

    def get_classes_stats(
        self,
        project_meta: Optional[ProjectMeta] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        if project_meta is None:
            project = PointcloudEpisodeProject(self.project_dir, OpenMode.READ)
            project_meta = project.meta
        class_items = {}
        class_objects = {}
        class_figures = {}
        for obj_class in project_meta.obj_classes:
            class_items[obj_class.name] = 0
            class_objects[obj_class.name] = 0
            class_figures[obj_class.name] = 0
        episode_ann: PointcloudEpisodeAnnotation = self.get_ann(project_meta)
        for ann_obj in episode_ann.objects:
            class_objects[ann_obj.obj_class.name] += 1
        for item_name in self:
            frame_index = self.get_frame_idx(item_name)
            item_figures = episode_ann.get_figures_on_frame(frame_index)
            item_class = {}
            for ptc_figure in item_figures:
                class_figures[ptc_figure.parent_object.obj_class.name] += 1
                item_class[ptc_figure.parent_object.obj_class.name] = True
            for obj_class in project_meta.obj_classes:
                if obj_class.name in item_class.keys():
                    class_items[obj_class.name] += 1

        result = {}
        if return_items_count:
            result["items_count"] = class_items
        if return_objects_count:
            result["objects_count"] = class_objects
        if return_figures_count:
            result["figures_count"] = class_figures
        return result

    def _add_ann_by_type(self, item_name, frame):
        if frame is None:
            self._item_to_ann[item_name] = ""
        elif isinstance(frame, int):
            self._item_to_ann[item_name] = str(frame)
        elif type(frame) is str:
            self._item_to_ann[item_name] = frame
        else:
            raise TypeError("Unsupported type {!r} for ann argument".format(type(frame)))

    def get_frame_idx(self, item_name: str) -> int:
        frame = self._item_to_ann.get(item_name, None)
        if frame is None:
            raise RuntimeError("Item {} not found in the project.".format(item_name))
        if self._item_to_ann[item_name] == "":
            return None
        return int(self._item_to_ann[item_name])


class PointcloudEpisodeProject(PointcloudProject):
    dataset_class = PointcloudEpisodeDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = PointcloudEpisodeDataset

    @classmethod
    def read_single(cls, dir) -> PointcloudEpisodeProject:
        return read_project_wrapper(dir, cls)

    def get_classes_stats(
        self,
        dataset_names: Optional[List[str]] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        return super(PointcloudEpisodeProject, self).get_classes_stats(
            dataset_names, return_objects_count, return_figures_count, return_items_count
        )

    @property
    def type(self) -> str:
        """
        Project type.

        :return: Project type.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            project = sly.PointcloudEpisodeProject("/home/admin/work/supervisely/projects/pointcloud_episode_project", sly.OpenMode.READ)
            print(project.type)
            # Output: 'point_cloud_episodes'
        """
        return ProjectType.POINT_CLOUD_EPISODES.value

    @staticmethod
    def get_train_val_splits_by_count(
        project_dir: str, train_count: int, val_count: int
    ) -> Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]:
        """
        Get train and val items information from project by given train and val counts.

        :param project_dir: Path to project directory.
        :type project_dir: str
        :param train_count: Number of train items.
        :type train_count: int
        :param val_count: Number of val items.
        :type val_count: int
        :raises: :class:`ValueError` if total_count != train_count + val_count
        :return: Tuple with lists of train items information and val items information
        :rtype: :class:`Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]`
        :Usage example:

         .. code-block:: python

            project_path = "/home/admin/work/supervisely/projects/pointcloud_project"
            project = sly.PointcloudEpisodeProject(project_path, sly.OpenMode.READ)
            train_count = 16
            val_count = 4
            train_items, val_items = project.get_train_val_splits_by_count(project_path, train_count, val_count)
        """

        def _list_items_for_splits(project) -> List[EpisodeItemInfo]:
            items = []
            for dataset in project.datasets:
                dataset: PointcloudEpisodeDataset
                for item_name in dataset:
                    items.append(
                        EpisodeItemInfo(
                            dataset_name=dataset.name,
                            name=item_name,
                            pointcloud_path=dataset.get_pointcloud_path(item_name),
                            related_images_dir=dataset.get_related_images_path(item_name),
                            frame_index=dataset.get_frame_idx(item_name),
                        )
                    )
            return items

        project = PointcloudEpisodeProject(project_dir, OpenMode.READ)
        if project.total_items != train_count + val_count:
            raise ValueError("total_count != train_count + val_count")
        all_items = _list_items_for_splits(project)
        random.shuffle(all_items)
        train_items = all_items[:train_count]
        val_items = all_items[train_count:]
        return train_items, val_items

    @staticmethod
    def get_train_val_splits_by_tag(
        project_dir: str,
        train_tag_name: str,
        val_tag_name: str,
        untagged: Optional[str] = "ignore",
    ) -> Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]:
        """
        Get train and val items information from project by given train and val tags names.

        :param project_dir: Path to project directory.
        :type project_dir: str
        :param train_tag_name: Train tag name.
        :type train_tag_name: str
        :param val_tag_name: Val tag name.
        :type val_tag_name: str
        :param untagged: Actions in case of absence of train_tag_name and val_tag_name in project.
        :type untagged: str, optional
        :raises: :class:`ValueError` if untagged not in ["ignore", "train", "val"]
        :return: Tuple with lists of train items information and val items information
        :rtype: :class:`Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            project_path = "/home/admin/work/supervisely/projects/pointcloud_project"
            project = sly.PointcloudEpisodeProject(project_path, sly.OpenMode.READ)
            train_tag_name = 'train'
            val_tag_name = 'val'
            train_items, val_items = project.get_train_val_splits_by_tag(project_path, train_tag_name, val_tag_name)
        """
        untagged_actions = ["ignore", "train", "val"]
        if untagged not in untagged_actions:
            raise ValueError(
                f"Unknown untagged action {untagged}. Should be one of {untagged_actions}"
            )
        project = PointcloudEpisodeProject(project_dir, OpenMode.READ)
        train_items = []
        val_items = []
        for dataset in project.datasets:
            ann = dataset.get_ann(project.meta)
            for item_name in dataset:
                item_paths = dataset.get_item_paths(item_name)
                frame_idx = dataset.get_frame_idx(item_name)

                info = EpisodeItemInfo(
                    dataset_name=dataset.name,
                    name=item_name,
                    pointcloud_path=item_paths.pointcloud_path,
                    related_images_dir=item_paths.related_images_dir,
                    frame_index=frame_idx,
                )
                frame_tags = ann.get_tags_on_frame(frame_idx)
                if frame_tags.get(train_tag_name) is not None:
                    train_items.append(info)
                if frame_tags.get(val_tag_name) is not None:
                    val_items.append(info)
                if frame_tags.get(train_tag_name) is None and frame_tags.get(val_tag_name) is None:
                    # untagged item
                    if untagged == "ignore":
                        continue
                    elif untagged == "train":
                        train_items.append(info)
                    elif untagged == "val":
                        val_items.append(info)
        return train_items, val_items

    @staticmethod
    def get_train_val_splits_by_dataset(
        project_dir: str, train_datasets: List[str], val_datasets: List[str]
    ) -> Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]:
        """
        Get train and val items information from project by given train and val datasets names.

        :param project_dir: Path to project directory.
        :type project_dir: str
        :param train_datasets: List of train datasets names.
        :type train_datasets: List[str]
        :param val_datasets: List of val datasets names.
        :type val_datasets: List[str]
        :raises: :class:`KeyError` if dataset name not found in project
        :return: Tuple with lists of train items information and val items information
        :rtype: :class:`Tuple[List[EpisodeItemInfo], List[EpisodeItemInfo]]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            project_path = "/home/admin/work/supervisely/projects/pointcloud_project"
            project = sly.PointcloudEpisodeProject(project_path, sly.OpenMode.READ)
            train_datasets = ['ds1', 'ds2']
            val_datasets = ['ds3', 'ds4']
            train_items, val_items = project.get_train_val_splits_by_dataset(project_path, train_datasets, val_datasets)
        """

        def _add_items_to_list(project, datasets_names, items_list):
            for dataset_name in datasets_names:
                dataset = project.datasets.get(dataset_name)
                if dataset is None:
                    raise KeyError(f"Dataset '{dataset_name}' not found")
                for item_name in dataset:
                    item_paths = dataset.get_item_paths(item_name)
                    frame_idx = dataset.get_frame_idx(item_name)
                    info = EpisodeItemInfo(
                        dataset_name=dataset.name,
                        name=item_name,
                        pointcloud_path=item_paths.pointcloud_path,
                        related_images_dir=item_paths.related_images_dir,
                        frame_index=frame_idx,
                    )
                    items_list.append(info)

        project = PointcloudEpisodeProject(project_dir, OpenMode.READ)
        train_items = []
        _add_items_to_list(project, train_datasets, train_items)
        val_items = []
        _add_items_to_list(project, val_datasets, val_items)
        return train_items, val_items

    @staticmethod
    def get_train_val_splits_by_collections(
        project_dir: str,
        train_collections: List[int],
        val_collections: List[int],
        project_id: int,
        api: Api,
    ) -> None:
        """
        Not available for PointcloudEpisodeProject class.
        :raises: :class:`NotImplementedError` in all cases.
        """
        raise NotImplementedError(
            f"Static method 'get_train_val_splits_by_collections()' is not supported for PointcloudEpisodeProject class now."
        )

    @staticmethod
    def download(
        api: Api,
        project_id: int,
        dest_dir: str,
        dataset_ids: Optional[List[int]] = None,
        download_pointclouds: Optional[bool] = True,
        download_related_images: Optional[bool] = True,
        download_pointclouds_info: Optional[bool] = False,
        batch_size: Optional[int] = 10,
        log_progress: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        **kwargs,
    ) -> None:
        """
        Download pointcloud episodes project from Supervisely to the given directory.

        :param api: Supervisely API address and token.
        :type api: :class:`Api<supervisely.api.api.Api>`
        :param project_id: Supervisely downloadable project ID.
        :type project_id: :class:`int`
        :param dest_dir: Destination directory.
        :type dest_dir: :class:`str`
        :param dataset_ids: Dataset IDs.
        :type dataset_ids: :class:`list` [ :class:`int` ], optional
        :param download_pointclouds: Download pointcloud data files or not.
        :type download_pointclouds: :class:`bool`, optional
        :param download_related_images: Download related images or not.
        :type download_related_images: :class:`bool`, optional
        :param download_pointclouds_info: Download pointcloud info .json files or not.
        :type download_pointclouds_info: :class:`bool`, optional
        :param batch_size: The number of images in the batch when they are loaded to a host.
        :type batch_size: :class:`int`, optional
        :param log_progress: Show uploading progress bar.
        :type log_progress: :class:`bool`
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: :class:`tqdm` or callable, optional
        :return: None
        :rtype: NoneType
        :Usage example:

        .. code-block:: python

                import supervisely as sly

                # Local destination project folder
                save_directory = "/home/admin/work/supervisely/source/ptc_project"

                # Obtain server address and your api_token from environment variables
                # Edit those values if you run this notebook on your own PC
                address = os.environ['SERVER_ADDRESS']
                token = os.environ['API_TOKEN']

                # Initialize API object
                api = sly.Api(address, token)
                project_id = 8888

                # Download Project
                sly.PointcloudEpisodeProject.download(api, project_id, save_directory)
                project_fs = sly.PointcloudEpisodeProject(save_directory, sly.OpenMode.READ)
        """
        download_pointcloud_episode_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            download_pcd=download_pointclouds,
            download_related_images=download_related_images,
            download_annotations=True,
            download_pointclouds_info=download_pointclouds_info,
            batch_size=batch_size,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )

    @staticmethod
    def upload(
        directory: str,
        api: Api,
        workspace_id: int,
        project_name: Optional[str] = None,
        log_progress: bool = True,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Tuple[int, str]:
        """
        Uploads pointcloud episodes project to Supervisely from the given directory.

        :param directory: Path to project directory.
        :type directory: str
        :param api: Supervisely API address and token.
        :type api: Api
        :param workspace_id: Workspace ID, where project will be uploaded.
        :type workspace_id: int
        :param project_name: Name of the project in Supervisely. Can be changed if project with the same name is already exists.
        :type project_name: str, optional
        :param log_progress: Show uploading progress bar.
        :type log_progress: bool
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Project ID and name. It is recommended to check that returned project name coincides with provided project name.
        :rtype: int, str
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            # Local folder with Pointcloud Project
            project_directory = "/home/admin/work/supervisely/source/episodes_project"

            # Obtain server address and your api_token from environment variables
            # Edit those values if you run this notebook on your own PC
            address = os.environ['SERVER_ADDRESS']
            token = os.environ['API_TOKEN']

            # Initialize API object
            api = sly.Api(address, token)

            # Upload Pointcloud Project
            project_id, project_name = sly.PointcloudEpisodeProject.upload(
                project_directory,
                api,
                workspace_id=45,
                project_name="My Episodes Project"
            )
        """
        return upload_pointcloud_episode_project(
            directory=directory,
            api=api,
            workspace_id=workspace_id,
            project_name=project_name,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )

    @staticmethod
    async def download_async(*args, **kwargs):
        raise NotImplementedError(
            f"Static method 'download_async()' is not supported for PointcloudEpisodeProject class now."
        )


def download_pointcloud_episode_project(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    download_pcd: Optional[bool] = True,
    download_related_images: Optional[bool] = True,
    download_annotations: Optional[bool] = True,
    download_pointclouds_info: Optional[bool] = False,
    batch_size: Optional[int] = 10,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> None:
    """
    Download pointcloud episode project to the local directory.

    :param api: Supervisely API address and token.
    :type api: Api
    :param project_id: Project ID to download.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded. Datasets could be downloaded from different projects but with the same data type.
    :type dataset_ids: list(int), optional
    :param download_pcd: Include pointcloud episode items in the download.
    :type download_pcd: bool, optional
    :param download_related_images: Include related context images in the download.
    :type download_related_images: bool, optional
    :param download_annotations: Include annotations in the download.
    :type download_annotations: bool, optional
    :param download_pointclouds_info: Include pointclouds info in the download.
    :type download_pointclouds_info: bool, optional
    :param batch_size: Size of a downloading batch.
    :type batch_size: int, optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool
    :param progress_cb: Function for tracking download progress.
    :type progress_cb: tqdm or callable, optional

    :return: None.
    :rtype: NoneType
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        from tqdm import tqdm
        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        dest_dir = 'your/local/dest/dir'

        # Download pointcloud episodes project
        project_id = 19636
        project_info = api.project.get_info_by_id(project_id)
        num_pointclouds_ep = project_info.items_count

        p = tqdm(desc="Downloading pointcloud project", total=num_pointclouds_ep)
        sly.download_pointcloud_project(
            api,
            project_id,
            dest_dir,
            progress_cb=p,
        )
    """

    key_id_map = KeyIdMap()
    project_fs = PointcloudEpisodeProject(dest_dir, OpenMode.CREATE)
    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    if progress_cb is not None:
        log_progress = False

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs: PointcloudEpisodeDataset = project_fs.create_dataset(dataset.name)
        pointclouds = api.pointcloud_episode.get_list(dataset.id)

        # Download annotation to project_path/dataset_path/annotation.json
        if download_annotations is True:
            ann_json = api.pointcloud_episode.annotation.download(dataset.id)
            annotation = dataset_fs.annotation_class.from_json(ann_json, meta, key_id_map)
            dataset_fs.set_ann(annotation)

        # frames --> pointcloud mapping to project_path/dataset_path/frame_pointcloud_map.json
        frame_name_map = api.pointcloud_episode.get_frame_name_map(dataset.id)
        frame_pointcloud_map_path = dataset_fs.get_frame_pointcloud_map_path()
        dump_json_file(frame_name_map, frame_pointcloud_map_path)

        ds_progress = progress_cb
        if log_progress:
            ds_progress = tqdm_sly(
                desc="Downloading episodes from: {!r}".format(dataset.name),
                total=len(pointclouds),
            )

        for batch in batched(pointclouds, batch_size=batch_size):
            pointcloud_ids = [pointcloud_info.id for pointcloud_info in batch]
            pointcloud_names = [pointcloud_info.name for pointcloud_info in batch]
            map_file_path = dataset_fs.get_frame_pointcloud_map_path()
            frame_to_pc_map = load_json_file(map_file_path)
            pc_to_frame = {v: k for k, v in frame_to_pc_map.items()}
            item_to_ann = {name: pc_to_frame[name] for name in pointcloud_names}

            batch_rimg_figures = {}
            if download_related_images:
                try:
                    rimgs = api.pointcloud_episode.get_list_related_images_batch(
                        dataset.id, pointcloud_ids
                    )
                    if len(rimgs) > 0:
                        rimg_ids = [rimg[ApiField.ID] for rimg in rimgs]
                        batch_rimg_figures = api.image.figure.download(
                            dataset_id=dataset.id, image_ids=rimg_ids
                        )
                    else:
                        batch_rimg_figures = []
                except Exception as e:
                    logger.info(
                        "INFO FOR DEBUGGING",
                        extra={
                            "project_id": project_id,
                            "dataset_id": dataset.id,
                            "pointcloud_ids": pointcloud_ids,
                        },
                    )
                    raise e

            for pointcloud_id, pointcloud_name, pointcloud_info in zip(
                pointcloud_ids, pointcloud_names, batch
            ):
                pointcloud_file_path = dataset_fs.generate_item_path(pointcloud_name)
                if download_pcd:
                    try:
                        api.pointcloud_episode.download_path(pointcloud_id, pointcloud_file_path)
                    except Exception as e:
                        logger.info(
                            "INFO FOR DEBUGGING",
                            extra={
                                "project_id": project_id,
                                "dataset_id": dataset.id,
                                "pointcloud_id": pointcloud_id,
                                "pointcloud_name": pointcloud_name,
                                "pointcloud_file_path": pointcloud_file_path,
                            },
                        )
                        raise e
                else:
                    touch(pointcloud_file_path)

                if download_related_images:
                    related_images_path = dataset_fs.get_related_images_path(pointcloud_name)
                    try:
                        related_images = api.pointcloud_episode.get_list_related_images(
                            pointcloud_id
                        )
                    except Exception as e:
                        logger.info(
                            "INFO FOR DEBUGGING",
                            extra={
                                "project_id": project_id,
                                "dataset_id": dataset.id,
                                "pointcloud_id": pointcloud_id,
                                "pointcloud_name": pointcloud_name,
                            },
                        )
                        raise e

                    for rimage_info in related_images:
                        name = rimage_info[ApiField.NAME]
                        if not sly_image.has_valid_ext(name):
                            new_name = get_file_name(name)  # to fix cases like .png.json
                            if sly_image.has_valid_ext(new_name):
                                name = new_name
                                rimage_info[ApiField.NAME] = name
                            else:
                                raise RuntimeError(
                                    "Something wrong with photo context filenames.\
                                                    Please, contact support"
                                )
                        rimage_id = rimage_info[ApiField.ID]

                        path_img = os.path.join(related_images_path, name)
                        path_json = os.path.join(related_images_path, name + ".json")
                        path_figures = os.path.join(related_images_path, name + ".figures.json")

                        try:
                            api.pointcloud_episode.download_related_image(rimage_id, path_img)
                        except Exception as e:
                            logger.info(
                                "INFO FOR DEBUGGING",
                                extra={
                                    "project_id": project_id,
                                    "dataset_id": dataset.id,
                                    "pointcloud_id": pointcloud_id,
                                    "pointcloud_name": pointcloud_name,
                                    "rimage_id": rimage_id,
                                    "path_img": path_img,
                                },
                            )
                            raise e
                        dump_json_file(rimage_info, path_json)

                        try:
                            if rimage_id in batch_rimg_figures:
                                rimg_figures = batch_rimg_figures[rimage_id]
                                rimg_figures_json = []
                                for fig in rimg_figures:
                                    fig_json = fig.to_json()
                                    if ApiField.OBJECT_ID in fig_json:
                                        fig_json[OBJECT_KEY] = str(
                                            key_id_map.get_object_key(fig_json[ApiField.OBJECT_ID])
                                        )
                                        fig_json.pop(ApiField.OBJECT_ID, None)
                                        rimg_figures_json.append(fig_json)
                                    else:
                                        raise RuntimeError(f"Figure {fig} has no object id")
                                dump_json_file(rimg_figures_json, path_figures)
                        except Exception as e:
                            logger.info(
                                "INFO FOR DEBUGGING",
                                extra={
                                    "project_id": project_id,
                                    "dataset_id": dataset.id,
                                    "pointcloud_id": pointcloud_id,
                                    "pointcloud_name": pointcloud_name,
                                    "rimage_id": rimage_id,
                                    "path_figures": path_figures,
                                },
                            )
                            raise e

                pointcloud_info = pointcloud_info._asdict() if download_pointclouds_info else None
                try:
                    dataset_fs.add_item_file(
                        pointcloud_name,
                        pointcloud_file_path,
                        item_to_ann[pointcloud_name],
                        _validate_item=False,
                        item_info=pointcloud_info,
                    )
                except Exception as e:
                    logger.info(
                        "INFO FOR DEBUGGING",
                        extra={
                            "project_id": project_id,
                            "dataset_id": dataset.id,
                            "pointcloud_id": pointcloud_id,
                            "pointcloud_name": pointcloud_name,
                            "pointcloud_file_path": pointcloud_file_path,
                            "item_info": pointcloud_info,
                        },
                    )
                    raise e

                if log_progress is True or progress_cb is not None:
                    ds_progress(1)

    project_fs.set_key_id_map(key_id_map)


def upload_pointcloud_episode_project(
    directory: str,
    api: Api,
    workspace_id: int,
    project_name: Optional[str] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> Tuple[int, str]:
    # STEP 0 — create project remotely
    project_fs = PointcloudEpisodeProject.read_single(directory)
    project_name = project_fs.name if project_name is None else project_name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project = api.project.create(workspace_id, project_name, ProjectType.POINT_CLOUD_EPISODES)
    api.project.update_meta(project.id, project_fs.meta.to_json())

    if progress_cb is not None:
        log_progress = False

    key_id_map = KeyIdMap()
    for dataset_fs in project_fs.datasets:
        dataset_fs: PointcloudEpisodeDataset
        ann_json_path = dataset_fs.get_ann_path()

        if os.path.isfile(ann_json_path):
            ann_json = load_json_file(ann_json_path)
            episode_annotation = PointcloudEpisodeAnnotation.from_json(ann_json, project_fs.meta)
        else:
            episode_annotation = PointcloudEpisodeAnnotation()

        dataset = api.dataset.create(
            project.id,
            dataset_fs.name,
            description=episode_annotation.description,
            change_name_if_conflict=True,
        )

        # STEP 1 — upload episodes
        items_infos = {"names": [], "paths": [], "metas": []}

        for item_name in dataset_fs:
            item_path, related_images_dir, frame_idx = dataset_fs.get_item_paths(item_name)

            item_meta = {"frame": frame_idx}

            items_infos["names"].append(item_name)
            items_infos["paths"].append(item_path)
            items_infos["metas"].append(item_meta)

        ds_progress = progress_cb
        if log_progress:
            ds_progress = tqdm_sly(
                desc="Uploading clouds to {!r}".format(dataset.name),
                total=len(dataset_fs),
            )
        try:
            pcl_infos = api.pointcloud_episode.upload_paths(
                dataset.id,
                names=items_infos["names"],
                paths=items_infos["paths"],
                metas=items_infos["metas"],
                progress_cb=ds_progress,
            )
        except Exception as e:
            logger.info(
                "INFO FOR DEBUGGING",
                extra={
                    "project_id": project.id,
                    "dataset_id": dataset.id,
                    "item_names": items_infos["names"],
                    "item_paths": items_infos["paths"],
                    "item_metas": items_infos["metas"],
                },
            )
            raise e
        # STEP 2 — upload annotations
        frame_to_pcl_ids = {pcl_info.frame: pcl_info.id for pcl_info in pcl_infos}
        try:
            api.pointcloud_episode.annotation.append(
                dataset.id, episode_annotation, frame_to_pcl_ids, key_id_map
            )
        except Exception as e:
            logger.info(
                "INFO FOR DEBUGGING",
                extra={
                    "project_id": project.id,
                    "dataset_id": dataset.id,
                    "frame_to_pcl_ids": frame_to_pcl_ids,
                    "ann": episode_annotation.to_json(),
                },
            )
            raise e

        # STEP 3 — upload photo context
        img_infos = {"img_paths": [], "img_metas": []}
        # STEP 3.1 — upload images
        pcl_to_rimg_figures: Dict[int, Dict[str, List[Dict]]] = {}
        pcl_to_hash_to_id: Dict[int, Dict[str, int]] = {}
        for pcl_info in pcl_infos:
            related_items = dataset_fs.get_related_images(pcl_info.name)
            images_paths_for_frame = [img_path for img_path, _ in related_items]

            img_infos["img_paths"].extend(images_paths_for_frame)

        rltd_progress = None
        if log_progress or progress_cb is not None:
            rltd_progress = tqdm_sly(
                desc="Uploading photo context to {!r}".format(dataset.name),
                total=len(img_infos["img_paths"]),
                leave=False,
            )
        try:
            images_hashes = api.pointcloud_episode.upload_related_images(
                img_infos["img_paths"],
                progress_cb=rltd_progress,
            )
        except Exception as e:
            logger.info(
                "INFO FOR DEBUGGING",
                extra={
                    "project_id": project.id,
                    "dataset_id": dataset.id,
                    "img_paths": img_infos["img_paths"],
                },
            )
            raise e

        # STEP 3.2 — upload images metas
        images_hashes_iterator = images_hashes.__iter__()
        for pcl_info in pcl_infos:
            related_items = dataset_fs.get_related_images(pcl_info.name)

            for img_ind, (_, meta_json) in enumerate(related_items):
                img_hash = next(images_hashes_iterator)
                if "deviceId" not in meta_json[ApiField.META].keys():
                    meta_json[ApiField.META]["deviceId"] = f"CAM_{str(img_ind).zfill(2)}"

                img_infos["img_metas"].append(
                    {
                        ApiField.ENTITY_ID: pcl_info.id,
                        ApiField.NAME: meta_json[ApiField.NAME],
                        ApiField.HASH: img_hash,
                        ApiField.META: meta_json[ApiField.META],
                    }
                )
                img_name = meta_json[ApiField.NAME]
                related_images_dir = dataset_fs.get_related_images_path(pcl_info.name)
                fig_json_path = os.path.join(related_images_dir, img_name + ".figures.json")
                if os.path.isfile(fig_json_path):
                    try:
                        figs_json = load_json_file(fig_json_path)
                        pcl_to_rimg_figures.setdefault(pcl_info.id, {})[img_hash] = figs_json
                    except Exception as e:
                        logger.info(
                            "INFO FOR DEBUGGING",
                            extra={
                                "project_id": project.id,
                                "dataset_id": dataset.id,
                                "pointcloud_id": pcl_info.id,
                                "fig_json_path": fig_json_path,
                            },
                        )
                        raise e

        if len(img_infos["img_metas"]) > 0:
            try:
                uploaded_rimgs = api.pointcloud_episode.add_related_images(img_infos["img_metas"])
                # build mapping hash->id
                for info, uploaded in zip(img_infos["img_metas"], uploaded_rimgs):
                    img_hash = info.get(ApiField.HASH)
                    img_id = (
                        uploaded.get(ApiField.ID)
                        if isinstance(uploaded, dict)
                        else getattr(uploaded, "id", None)
                    )
                    if img_hash is not None and img_id is not None:
                        pcl_to_hash_to_id[img_hash] = img_id
            except Exception as e:
                logger.info(
                    "INFO FOR DEBUGGING",
                    extra={
                        "project_id": project.id,
                        "dataset_id": dataset.id,
                        "rimg_infos": img_infos["img_metas"],
                    },
                )
                raise e

        for pcl_info in pcl_infos:
            rimg_figures = pcl_to_rimg_figures.get(pcl_info.id)
            if not rimg_figures:
                continue

            try:
                for img_hash, figs_json in rimg_figures.items():
                    if img_hash in pcl_to_hash_to_id:
                        img_id = pcl_to_hash_to_id[img_hash]
                        for fig in figs_json:
                            fig[ApiField.ENTITY_ID] = img_id
                            fig[ApiField.DATASET_ID] = dataset.id
                            fig[ApiField.PROJECT_ID] = project.id
                            fig[ApiField.OBJECT_ID] = key_id_map.get_object_id(
                                UUID(fig[OBJECT_KEY])
                            )

                api.image.figure.create_bulk(
                    figures_json=[fig for figs in rimg_figures.values() for fig in figs],
                    dataset_id=dataset.id,
                )

            except Exception as e:
                logger.info(
                    "INFO FOR DEBUGGING",
                    extra={
                        "project_id": project.id,
                        "dataset_id": dataset.id,
                        "pointcloud_id": pcl_info.id,
                        "pointcloud_name": pcl_info.name,
                        "rimg_figures": rimg_figures,
                    },
                )
                raise e

    return project.id, project.name




from __future__ import absolute_import, division

import warnings
from supervisely import logger

try:
    import torch  # pylint: disable=import-error
    from torch import nn
    from torch.nn import functional as F
except ImportError:
    logger.warning("torch is not installed, OSNet re-ID cannot be used.")
    

__all__ = ["osnet_x1_0", "osnet_x0_75", "osnet_x0_5", "osnet_x0_25", "osnet_ibn_x1_0"]

pretrained_urls = {
    "osnet_x1_0": "https://drive.google.com/uc?id=1LaG1EJpHrxdAxKnSCJ_i0u-nbxSAeiFY",
    "osnet_x0_75": "https://drive.google.com/uc?id=1uwA9fElHOk3ZogwbeY5GkLI6QPTX70Hq",
    "osnet_x0_5": "https://drive.google.com/uc?id=16DGLbZukvVYgINws8u8deSaOqjybZ83i",
    "osnet_x0_25": "https://drive.google.com/uc?id=1rb8UN5ZzPKRc_xvtHlyDh-cSz88YX9hs",
    "osnet_ibn_x1_0": "https://drive.google.com/uc?id=1sr90V6irlYYDd4_4ISU2iruoRG8J__6l",
}


##########
# Basic layers
##########
class ConvLayer(nn.Module):
    """Convolution layer (conv + bn + relu)."""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        groups=1,
        IN=False,
    ):
        super(ConvLayer, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=False,
            groups=groups,
        )
        if IN:
            self.bn = nn.InstanceNorm2d(out_channels, affine=True)
        else:
            self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Conv1x1(nn.Module):
    """1x1 convolution + bn + relu."""

    def __init__(self, in_channels, out_channels, stride=1, groups=1):
        super(Conv1x1, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            1,
            stride=stride,
            padding=0,
            bias=False,
            groups=groups,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Conv1x1Linear(nn.Module):
    """1x1 convolution + bn (w/o non-linearity)."""

    def __init__(self, in_channels, out_channels, stride=1):
        super(Conv1x1Linear, self).__init__()
        self.conv = nn.Conv2d(
            in_channels, out_channels, 1, stride=stride, padding=0, bias=False
        )
        self.bn = nn.BatchNorm2d(out_channels)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x


class Conv3x3(nn.Module):
    """3x3 convolution + bn + relu."""

    def __init__(self, in_channels, out_channels, stride=1, groups=1):
        super(Conv3x3, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            3,
            stride=stride,
            padding=1,
            bias=False,
            groups=groups,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class LightConv3x3(nn.Module):
    """Lightweight 3x3 convolution.

    1x1 (linear) + dw 3x3 (nonlinear).
    """

    def __init__(self, in_channels, out_channels):
        super(LightConv3x3, self).__init__()
        self.conv1 = nn.Conv2d(
            in_channels, out_channels, 1, stride=1, padding=0, bias=False
        )
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            3,
            stride=1,
            padding=1,
            bias=False,
            groups=out_channels,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


##########
# Building blocks for omni-scale feature learning
##########
class ChannelGate(nn.Module):
    """A mini-network that generates channel-wise gates conditioned on input tensor."""

    def __init__(
        self,
        in_channels,
        num_gates=None,
        return_gates=False,
        gate_activation="sigmoid",
        reduction=16,
        layer_norm=False,
    ):
        super(ChannelGate, self).__init__()
        if num_gates is None:
            num_gates = in_channels
        self.return_gates = return_gates
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc1 = nn.Conv2d(
            in_channels, in_channels // reduction, kernel_size=1, bias=True, padding=0
        )
        self.norm1 = None
        if layer_norm:
            self.norm1 = nn.LayerNorm((in_channels // reduction, 1, 1))
        self.relu = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv2d(
            in_channels // reduction, num_gates, kernel_size=1, bias=True, padding=0
        )
        if gate_activation == "sigmoid":
            self.gate_activation = nn.Sigmoid()
        elif gate_activation == "relu":
            self.gate_activation = nn.ReLU(inplace=True)
        elif gate_activation == "linear":
            self.gate_activation = None
        else:
            raise RuntimeError("Unknown gate activation: {}".format(gate_activation))

    def forward(self, x):
        input = x
        x = self.global_avgpool(x)
        x = self.fc1(x)
        if self.norm1 is not None:
            x = self.norm1(x)
        x = self.relu(x)
        x = self.fc2(x)
        if self.gate_activation is not None:
            x = self.gate_activation(x)
        if self.return_gates:
            return x
        return input * x


class OSBlock(nn.Module):
    """Omni-scale feature learning block."""

    def __init__(
        self, in_channels, out_channels, IN=False, bottleneck_reduction=4, **kwargs
    ):
        super(OSBlock, self).__init__()
        mid_channels = out_channels // bottleneck_reduction
        self.conv1 = Conv1x1(in_channels, mid_channels)
        self.conv2a = LightConv3x3(mid_channels, mid_channels)
        self.conv2b = nn.Sequential(
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
        )
        self.conv2c = nn.Sequential(
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
        )
        self.conv2d = nn.Sequential(
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
            LightConv3x3(mid_channels, mid_channels),
        )
        self.gate = ChannelGate(mid_channels)
        self.conv3 = Conv1x1Linear(mid_channels, out_channels)
        self.downsample = None
        if in_channels != out_channels:
            self.downsample = Conv1x1Linear(in_channels, out_channels)
        self.IN = None
        if IN:
            self.IN = nn.InstanceNorm2d(out_channels, affine=True)

    def forward(self, x):
        identity = x
        x1 = self.conv1(x)
        x2a = self.conv2a(x1)
        x2b = self.conv2b(x1)
        x2c = self.conv2c(x1)
        x2d = self.conv2d(x1)
        x2 = self.gate(x2a) + self.gate(x2b) + self.gate(x2c) + self.gate(x2d)
        x3 = self.conv3(x2)
        if self.downsample is not None:
            identity = self.downsample(identity)
        out = x3 + identity
        if self.IN is not None:
            out = self.IN(out)
        return F.relu(out)


##########
# Network architecture
##########
class OSNet(nn.Module):
    """Omni-Scale Network.

    Reference:
        - Zhou et al. Omni-Scale Feature Learning for Person Re-Identification. ICCV, 2019.
        - Zhou et al. Learning Generalisable Omni-Scale Representations
          for Person Re-Identification. TPAMI, 2021.
    """

    def __init__(
        self,
        num_classes,
        blocks,
        layers,
        channels,
        feature_dim=512,
        loss="softmax",
        IN=False,
        **kwargs,
    ):
        super(OSNet, self).__init__()
        num_blocks = len(blocks)
        assert num_blocks == len(layers)
        assert num_blocks == len(channels) - 1
        self.loss = loss
        self.feature_dim = feature_dim

        # convolutional backbone
        self.conv1 = ConvLayer(3, channels[0], 7, stride=2, padding=3, IN=IN)
        self.maxpool = nn.MaxPool2d(3, stride=2, padding=1)
        self.conv2 = self._make_layer(
            blocks[0],
            layers[0],
            channels[0],
            channels[1],
            reduce_spatial_size=True,
            IN=IN,
        )
        self.conv3 = self._make_layer(
            blocks[1], layers[1], channels[1], channels[2], reduce_spatial_size=True
        )
        self.conv4 = self._make_layer(
            blocks[2], layers[2], channels[2], channels[3], reduce_spatial_size=False
        )
        self.conv5 = Conv1x1(channels[3], channels[3])
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        # fully connected layer
        self.fc = self._construct_fc_layer(
            self.feature_dim, channels[3], dropout_p=None
        )
        # identity classification layer
        self.classifier = nn.Linear(self.feature_dim, num_classes)

        self._init_params()

    def _make_layer(
        self, block, layer, in_channels, out_channels, reduce_spatial_size, IN=False
    ):
        layers = []

        layers.append(block(in_channels, out_channels, IN=IN))
        for i in range(1, layer):
            layers.append(block(out_channels, out_channels, IN=IN))

        if reduce_spatial_size:
            layers.append(
                nn.Sequential(
                    Conv1x1(out_channels, out_channels), nn.AvgPool2d(2, stride=2)
                )
            )

        return nn.Sequential(*layers)

    def _construct_fc_layer(self, fc_dims, input_dim, dropout_p=None):
        if fc_dims is None or fc_dims < 0:
            self.feature_dim = input_dim
            return None

        if isinstance(fc_dims, int):
            fc_dims = [fc_dims]

        layers = []
        for dim in fc_dims:
            layers.append(nn.Linear(input_dim, dim))
            layers.append(nn.BatchNorm1d(dim))
            layers.append(nn.ReLU(inplace=True))
            if dropout_p is not None:
                layers.append(nn.Dropout(p=dropout_p))
            input_dim = dim

        self.feature_dim = fc_dims[-1]

        return nn.Sequential(*layers)

    def _init_params(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def featuremaps(self, x):
        x = self.conv1(x)
        x = self.maxpool(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        return x

    def forward(self, x, return_featuremaps=False):
        x = self.featuremaps(x)
        if return_featuremaps:
            return x
        v = self.global_avgpool(x)
        v = v.view(v.size(0), -1)
        if self.fc is not None:
            v = self.fc(v)
        if not self.training:
            return v
        y = self.classifier(v)
        if self.loss == "softmax":
            return y
        elif self.loss == "triplet":
            return y, v
        else:
            raise KeyError("Unsupported loss: {}".format(self.loss))


def init_pretrained_weights(model, key=""):
    """Initializes model with pretrained weights.

    Layers that don't match with pretrained layers in name or size are kept unchanged.
    """
    import errno
    import os
    from collections import OrderedDict

    import gdown  # pylint: disable=import-error

    def _get_torch_home():
        ENV_TORCH_HOME = "TORCH_HOME"
        ENV_XDG_CACHE_HOME = "XDG_CACHE_HOME"
        DEFAULT_CACHE_DIR = "~/.cache"
        torch_home = os.path.expanduser(
            os.getenv(
                ENV_TORCH_HOME,
                os.path.join(os.getenv(ENV_XDG_CACHE_HOME, DEFAULT_CACHE_DIR), "torch"),
            )
        )
        return torch_home

    torch_home = _get_torch_home()
    model_dir = os.path.join(torch_home, "checkpoints")
    try:
        os.makedirs(model_dir)
    except OSError as e:
        if e.errno == errno.EEXIST:
            # Directory already exists, ignore.
            pass
        else:
            # Unexpected OSError, re-raise.
            raise
    filename = key + "_imagenet.pth"
    cached_file = os.path.join(model_dir, filename)

    if not os.path.exists(cached_file):
        gdown.download(pretrained_urls[key], cached_file, quiet=False)

    state_dict = torch.load(cached_file)
    model_dict = model.state_dict()
    new_state_dict = OrderedDict()
    matched_layers, discarded_layers = [], []

    for k, v in state_dict.items():
        if k.startswith("module."):
            k = k[7:]  # discard module.

        if k in model_dict and model_dict[k].size() == v.size():
            new_state_dict[k] = v
            matched_layers.append(k)
        else:
            discarded_layers.append(k)

    model_dict.update(new_state_dict)
    model.load_state_dict(model_dict)

    if len(matched_layers) == 0:
        warnings.warn(
            'The pretrained weights from "{}" cannot be loaded, '
            "please check the key names manually "
            "(** ignored and continue **)".format(cached_file)
        )
    else:
        print(
            'Successfully loaded imagenet pretrained weights from "{}"'.format(
                cached_file
            )
        )
        if len(discarded_layers) > 0:
            print(
                "** The following layers are discarded "
                "due to unmatched keys or layer size: {}".format(discarded_layers)
            )


##########
# Instantiation
##########
def osnet_x1_0(num_classes=1000, pretrained=True, loss="softmax", **kwargs):
    # standard size (width x1.0)
    model = OSNet(
        num_classes,
        blocks=[OSBlock, OSBlock, OSBlock],
        layers=[2, 2, 2],
        channels=[64, 256, 384, 512],
        loss=loss,
        **kwargs,
    )
    if pretrained:
        init_pretrained_weights(model, key="osnet_x1_0")
    return model




def osnet_x0_75(num_classes=1000, pretrained=True, loss="softmax", **kwargs):
    # medium size (width x0.75)
    model = OSNet(
        num_classes,
        blocks=[OSBlock, OSBlock, OSBlock],
        layers=[2, 2, 2],
        channels=[48, 192, 288, 384],
        loss=loss,
        **kwargs,
    )
    if pretrained:
        init_pretrained_weights(model, key="osnet_x0_75")
    return model


def osnet_x0_5(num_classes=1000, pretrained=True, loss="softmax", **kwargs):
    # tiny size (width x0.5)
    model = OSNet(
        num_classes,
        blocks=[OSBlock, OSBlock, OSBlock],
        layers=[2, 2, 2],
        channels=[32, 128, 192, 256],
        loss=loss,
        **kwargs,
    )
    if pretrained:
        init_pretrained_weights(model, key="osnet_x0_5")
    return model


def osnet_x0_25(num_classes=1000, pretrained=True, loss="softmax", **kwargs):
    # very tiny size (width x0.25)
    model = OSNet(
        num_classes,
        blocks=[OSBlock, OSBlock, OSBlock],
        layers=[2, 2, 2],
        channels=[16, 64, 96, 128],
        loss=loss,
        **kwargs,
    )
    if pretrained:
        init_pretrained_weights(model, key="osnet_x0_25")
    return model


def osnet_ibn_x1_0(num_classes=1000, pretrained=True, loss="softmax", **kwargs):
    # standard size (width x1.0) + IBN layer
    # Ref: Pan et al. Two at Once: Enhancing Learning and Generalization Capacities via IBN-Net. ECCV, 2018.
    model = OSNet(
        num_classes,
        blocks=[OSBlock, OSBlock, OSBlock],
        layers=[2, 2, 2],
        channels=[64, 256, 384, 512],
        loss=loss,
        IN=True,
        **kwargs,
    )
    if pretrained:
        init_pretrained_weights(model, key="osnet_ibn_x1_0")
    return model



from typing import Dict, List, Union

import cv2
import numpy as np

from supervisely._utils import batched_iter
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta import TagApplicableTo, TagMeta, TagValueType
from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.image_api import ImageInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.api.video.video_api import VideoInfo
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import tqdm_sly
from supervisely.video.video import VideoFrameReader
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_annotation import VideoAnnotation

VIDEO_OBJECT_TAG_META = TagMeta(
    "object_id",
    value_type=TagValueType.ANY_NUMBER,
    applicable_to=TagApplicableTo.OBJECTS_ONLY,
)
AUTO_TRACKED_TAG_META = TagMeta(
    "auto-tracked", TagValueType.NONE, applicable_to=TagApplicableTo.OBJECTS_ONLY
)


class ApiContext:
    def __init__(self):
        self.project_info: Dict[int, ProjectInfo] = {}
        self.project_meta: Dict[int, ProjectMeta] = {}
        self.dataset_info: Dict[int, DatasetInfo] = {}
        self.video_info: Dict[int, VideoInfo] = {}
        self.children_datasets: Dict[int, List[DatasetInfo]] = {}
        self.children_items: Dict[int, List[Union[ImageInfo, VideoInfo]]] = {}

    def __getattr__(self, item: str) -> Dict:
        if not hasattr(self, item):
            new_dict = {}
            setattr(self, item, new_dict)
        return getattr(self, item)


class SamplingSettings:
    ONLY_ANNOTATED = "annotated"
    START = "start"
    END = "end"
    STEP = "step"
    RESIZE = "resize"
    COPY_ANNOTATIONS = "copy_annotations"


def _get_frame_indices(frames_count, start, end, step, only_annotated, video_annotation):
    frame_indices = list(range(start, end if end is not None else frames_count))
    if only_annotated:
        annotated_frames = set()
        for frame in video_annotation.frames:
            frame: Frame
            if frame.figures:
                annotated_frames.add(frame.index)
        frame_indices = [idx for idx in frame_indices if idx in annotated_frames]
    frame_indices = [frame_indices[i] for i in range(0, len(frame_indices), step)]
    return frame_indices


def _frame_to_annotation(frame: Frame, video_annotation: VideoAnnotation) -> Annotation:

    labels = []
    for figure in frame.figures:
        tags = []
        video_object = figure.parent_object
        obj_class = video_object.obj_class
        geometry = figure.geometry
        for tag in video_object.tags:
            if tag.frame_range is None or tag.frame_range[0] <= frame.index <= tag.frame_range[1]:
                tags.append(Tag(tag.meta, tag.value, labeler_login=tag.labeler_login))
        tags.append(Tag(VIDEO_OBJECT_TAG_META, video_object.class_id))
        if figure.track_id is not None:
            tags.append(Tag(AUTO_TRACKED_TAG_META, None, labeler_login=video_object.labeler_login))
        label = Label(geometry, obj_class, TagCollection(tags))
        labels.append(label)
    img_tags = []
    for tag in video_annotation.tags:
        if tag.frame_range is None or tag.frame_range[0] <= frame.index <= tag.frame_range[1]:
            img_tags.append(Tag(tag.meta, tag.value, labeler_login=tag.labeler_login))
    return Annotation(video_annotation.img_size, labels=labels, img_tags=TagCollection(img_tags))


def _upload_annotations(api: Api, image_ids, frame_indices, video_annotation: VideoAnnotation):
    anns = []
    for image_id, frame_index in zip(image_ids, frame_indices):
        frame = video_annotation.frames.get(frame_index, None)
        if frame is not None:
            anns.append(_frame_to_annotation(frame, video_annotation))
    api.annotation.upload_anns(image_ids, anns=anns)


def _upload_frames(
    api: Api,
    frames: List[np.ndarray],
    video_name: str,
    video_frames_count: int,
    indices: List[int],
    dataset_id: int,
    sample_info: Dict = None,
    context: ApiContext = None,
    copy_annotations: bool = False,
    video_annotation: VideoAnnotation = None,
) -> List[int]:
    if sample_info is None:
        sample_info = {}
    if context is not None:
        if dataset_id not in context.children_items:
            context.children_items[dataset_id] = api.image.get_list(dataset_id)
        existing_images = context.children_items[dataset_id]
    else:
        existing_images = api.image.get_list(dataset_id)

    name_to_info = {image.name: image for image in existing_images}

    image_ids = [None] * len(frames)
    to_upload = []
    for i, index in enumerate(indices):
        digits = len(str(video_frames_count))
        image_name = f"{video_name}_frame_{str(index).zfill(digits)}.png"
        if image_name in name_to_info:
            image_ids[i] = name_to_info[image_name].id
        else:
            to_upload.append((image_name, i))

    if to_upload:
        frames = [frames[i] for _, i in to_upload]
        names = [name for name, _ in to_upload]
        metas = [{**sample_info, "frame_index": i} for _, i in to_upload]
        uploaded = api.image.upload_nps(
            dataset_id=dataset_id,
            names=names,
            imgs=frames,
            metas=metas,
        )

        for image_info, (_, i) in zip(uploaded, to_upload):
            image_ids[i] = image_info.id

    if copy_annotations:
        _upload_annotations(api, image_ids, indices, video_annotation)

    return image_ids


def sample_video(
    api: Api,
    video_id: int,
    dst_dataset_info: DatasetInfo,
    settings: Dict,
    sample_info: Dict = None,
    context: ApiContext = None,
    progress: tqdm_sly = None,
):
    dst_parent_info = dst_dataset_info
    only_annotated = settings.get(SamplingSettings.ONLY_ANNOTATED, False)
    start_frame = settings.get(SamplingSettings.START, 0)
    end_frame = settings.get(SamplingSettings.END, None)
    step = settings.get(SamplingSettings.STEP, 1)
    resize = settings.get(SamplingSettings.RESIZE, None)
    copy_annotations = settings.get(SamplingSettings.COPY_ANNOTATIONS, False)

    if context is None:
        context = ApiContext()
    if video_id not in context.video_info:
        context.video_info[video_id] = api.video.get_info_by_id(video_id)
    video_info = context.video_info[video_id]

    project_id = video_info.project_id
    if project_id not in context.project_meta:
        context.project_meta[project_id] = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_meta = context.project_meta[project_id]

    video_annotation = VideoAnnotation.from_json(
        api.video.annotation.download(video_info.id), project_meta, key_id_map=KeyIdMap()
    )

    progress_cb = None
    if progress is not None:
        size = int(video_info.file_meta["size"])
        progress.reset(size)
        progress.unit = "B"
        progress.unit_scale = True
        progress.unit_divisor = 1024
        progress.message = f"Downloading {video_info.name} [{video_info.id}]"
        progress.desc = progress.message
        progress.refresh()
        progress_cb = progress.update

    video_path = f"/tmp/{video_info.name}"
    api.video.download_path(video_info.id, video_path, progress_cb=progress_cb)

    frame_indices = _get_frame_indices(
        video_info.frames_count, start_frame, end_frame, step, only_annotated, video_annotation
    )

    dst_dataset_info = _get_or_create_dst_dataset(
        api=api,
        src_info=video_info,
        dst_parent_info=dst_parent_info,
        sample_info=sample_info,
        context=context,
    )

    if progress is not None:
        progress.reset(len(frame_indices))
        progress.unit = "it"
        progress.unit_scale = False
        progress.unit_divisor = 1000
        progress.message = f"Processing {video_info.name} [{video_info.id}]"
        progress.desc = progress.message
        progress.miniters = 1
        progress.refresh()

    with VideoFrameReader(video_path, frame_indices) as reader:
        for batch in batched_iter(zip(reader, frame_indices), 10):
            frames, indices = zip(*batch)
            for frame in frames:
                if resize:
                    cv2.resize(frame, [*resize, frame.shape[2]], interpolation=cv2.INTER_LINEAR)

            image_ids = _upload_frames(
                api=api,
                frames=frames,
                video_name=video_info.name,
                video_frames_count=video_info.frames_count,
                indices=indices,
                dataset_id=dst_dataset_info.id,
                sample_info=sample_info,
                context=context,
                copy_annotations=copy_annotations,
                video_annotation=video_annotation,
            )

            if progress is not None:
                progress.update(len(image_ids))


def _get_or_create_dst_dataset(
    api: Api,
    src_info: Union[DatasetInfo, VideoInfo],
    dst_parent_info: Union[ProjectInfo, DatasetInfo],
    sample_info: Dict = None,
    context: ApiContext = None,
) -> DatasetInfo:
    if isinstance(dst_parent_info, ProjectInfo):
        parent = None
        project_id = dst_parent_info.id
    else:
        parent = dst_parent_info.id
        project_id = dst_parent_info.project_id

    if context is not None:
        if dst_parent_info.id not in context.children_datasets:
            context.children_datasets[dst_parent_info.id] = api.dataset.get_list(
                project_id, parent_id=parent
            )
        dst_datasets = context.children_datasets[dst_parent_info.id]
    else:
        dst_datasets = api.dataset.get_list(project_id, parent_id=parent)

    for dst_dataset in dst_datasets:
        if dst_dataset.name == src_info.name:
            return dst_dataset

    if isinstance(src_info, DatasetInfo):
        src_dataset_info = src_info
        description = (
            f"Sample dataset made from project #{src_info.parent_id}, dataset #{src_info.id}"
        )
    else:
        if context is not None:
            if src_info.dataset_id not in context.dataset_info:
                context.dataset_info[src_info.dataset_id] = api.dataset.get_info_by_id(
                    src_info.dataset_id
                )
            src_dataset_info = context.dataset_info[src_info.dataset_id]
        else:
            src_dataset_info = api.dataset.get_info_by_id(src_info.dataset_id)
        description = f"Sample dataset made from project #{src_info.project_id}, dataset #{src_info.dataset_id}, video #{src_info.id}"

    dst_dataset = api.dataset.create(
        project_id,
        name=src_info.name,
        description=description,
        parent_id=parent,
    )
    if sample_info is None:
        if context is not None:
            if src_info.project_id not in context.project_info:
                context.project_info[src_info.project_id] = api.project.get_info_by_id(
                    src_info.project_id
                )
            src_project_info = context.project_info[src_info.project_id]
        else:
            src_project_info = api.project.get_info_by_id(src_info.project_id)
        sample_info = {
            "is_sample": True,
            "video_project_id": src_info.project_id,
            "video_project_name": src_project_info.name,
        }
    sample_info.update(
        {
            "video_dataset_id": src_dataset_info.id,
            "video_dataset_name": src_dataset_info.name,
        }
    )
    if isinstance(src_info, VideoInfo):
        sample_info.update(
            {
                "video_id": src_info.id,
                "video_name": src_info.name,
            }
        )
    api.dataset.update_custom_data(
        dst_dataset.id,
        custom_data=sample_info,
    )
    if context is not None:
        context.dataset_info[dst_dataset.id] = dst_dataset
    return dst_dataset


def sample_video_dataset(
    api: Api,
    src_dataset_id: int,
    dst_parent_info: Union[ProjectInfo, DatasetInfo],
    settings: Dict,
    sample_info: Dict = None,
    context: ApiContext = None,
    datasets_ids_whitelist: List[int] = None,
    items_progress_cb: tqdm_sly = None,
    video_progress: tqdm_sly = None,
):
    if context is None:
        context = ApiContext()

    if not (
        datasets_ids_whitelist is None
        or src_dataset_id in datasets_ids_whitelist  # this dataset should be sampled
        or _has_children_datasets(
            api, src_dataset_id, context=context, children_ids=datasets_ids_whitelist
        )  # has children datasets that should be sampled
    ):
        return None

    src_dataset_info = context.dataset_info.get(src_dataset_id, None)
    if src_dataset_info is None:
        src_dataset_info = api.dataset.get_info_by_id(src_dataset_id)
        context.dataset_info[src_dataset_id] = src_dataset_info

    dst_dataset = _get_or_create_dst_dataset(
        api=api,
        src_info=src_dataset_info,
        dst_parent_info=dst_parent_info,
        sample_info=sample_info,
        context=context,
    )

    if datasets_ids_whitelist is None or src_dataset_id in datasets_ids_whitelist:
        video_infos = api.video.get_list(src_dataset_id)
        for video_info in video_infos:
            sample_video(
                api=api,
                video_id=video_info.id,
                dst_dataset_info=dst_dataset,
                settings=settings,
                sample_info=sample_info.copy(),
                context=context,
                progress=video_progress,
            )
            if items_progress_cb is not None:
                items_progress_cb()

    if src_dataset_id not in context.children_datasets:
        if src_dataset_id not in context.dataset_info:
            context.dataset_info[src_dataset_id] = api.dataset.get_info_by_id(src_dataset_id)
        src_dataset_info = context.dataset_info[src_dataset_id]
        context.children_datasets[src_dataset_id] = api.dataset.get_list(
            src_dataset_info.project_id, parent_id=src_dataset_info.id
        )
    datasets = context.children_datasets[src_dataset_id]
    for dataset in datasets:
        sample_video_dataset(
            api=api,
            src_dataset_id=dataset.id,
            dst_parent_info=dst_dataset,
            settings=settings,
            datasets_ids_whitelist=datasets_ids_whitelist,
            sample_info=sample_info,
            context=context,
            items_progress_cb=items_progress_cb,
            video_progress=video_progress,
        )
    return dst_dataset


def _update_meta(
    api: Api, src_project_meta: ProjectMeta, dst_project_id: int, context: ApiContext = None
):
    if context is None:
        context = ApiContext()
    if dst_project_id not in context.project_meta:
        context.project_meta[dst_project_id] = ProjectMeta.from_json(
            api.project.get_meta(dst_project_id)
        )
    dst_project_meta = context.project_meta[dst_project_id]
    dst_project_meta = dst_project_meta.merge(src_project_meta)
    if dst_project_meta.get_tag_meta(VIDEO_OBJECT_TAG_META.name) is None:
        dst_project_meta = dst_project_meta.add_tag_meta(VIDEO_OBJECT_TAG_META)
    if dst_project_meta.get_tag_meta(AUTO_TRACKED_TAG_META.name) is None:
        dst_project_meta = dst_project_meta.add_tag_meta(AUTO_TRACKED_TAG_META)

    if dst_project_meta != src_project_meta:
        dst_project_meta = api.project.update_meta(dst_project_id, dst_project_meta.to_json())
        context.project_meta[dst_project_id] = dst_project_meta


def _get_or_create_dst_project(
    api: Api,
    src_project_id: int,
    dst_project_id: Union[int, None] = None,
    sample_info: Dict = None,
    context: ApiContext = None,
) -> ProjectInfo:
    if dst_project_id is None:
        # get source project info
        if context is None:
            src_project_info = api.project.get_info_by_id(src_project_id)
        else:
            if src_project_id not in context.project_info:
                context.project_info[src_project_id] = api.project.get_info_by_id(src_project_id)
            src_project_info = context.project_info[src_project_id]
        # create new project
        if sample_info is None:
            sample_info = {}
        sample_info.update(
            {
                "is_sample": True,
                "video_project_id": src_project_id,
                "video_project_name": src_project_info.name,
            }
        )
        dst_project = api.project.create(
            src_project_info.workspace_id,
            f"{src_project_info.name}(images)",
            description=f"Sample project made from project #{src_project_info.id}",
            change_name_if_conflict=True,
        )
        api.project.update_custom_data(dst_project.id, sample_info)
    else:
        # use existing project
        dst_project = api.project.get_info_by_id(dst_project_id)
    if context is not None:
        context.project_info[dst_project.id] = dst_project
    if src_project_id not in context.project_meta:
        context.project_meta[src_project_id] = ProjectMeta.from_json(
            api.project.get_meta(src_project_id)
        )
    src_project_meta = context.project_meta[src_project_id]
    _update_meta(api, src_project_meta, dst_project.id, context=context)
    return dst_project


def _has_children_datasets(
    api: Api, dataset_id: int, context: ApiContext = None, children_ids: List[int] = None
) -> bool:
    if context is None:
        context = ApiContext()

    if dataset_id not in context.dataset_info:
        context.dataset_info[dataset_id] = api.dataset.get_info_by_id(dataset_id)
    dataset_info = context.dataset_info[dataset_id]
    if dataset_id not in context.children_datasets:
        context.children_datasets[dataset_id] = api.dataset.get_list(
            project_id=dataset_info.project_id, parent_id=dataset_id
        )

    if children_ids is None:
        return len(context.children_datasets[dataset_id]) > 0
    for child_dataset in context.children_datasets[dataset_id]:
        if child_dataset.id in children_ids:
            return True
        if _has_children_datasets(
            api, child_dataset.id, context=context, children_ids=children_ids
        ):
            return True
    return False


def sample_video_project(
    api: Api,
    project_id: int,
    settings: Dict,
    dst_project_id: Union[int, None] = None,
    datasets_ids: List[int] = None,
    context: ApiContext = None,
    items_progress_cb: tqdm_sly = None,
    video_progress: tqdm_sly = None,
):
    if context is None:
        context = ApiContext()

    if project_id not in context.project_info:
        context.project_info[project_id] = api.project.get_info_by_id(project_id)

    src_project_info = context.project_info[project_id]

    sample_info = {
        "is_sample": True,
        "video_project_id": src_project_info.id,
        "video_project_name": src_project_info.name,
    }
    dst_project_info = _get_or_create_dst_project(
        api, project_id, dst_project_id, sample_info, context
    )

    # non recursive. Nested datasets are handled by sample_video_dataset
    if project_id not in context.children_datasets:
        datasets = api.dataset.get_list(project_id)
        context.children_datasets[project_id] = datasets

    dataset_infos = context.children_datasets[project_id]
    for dataset_info in dataset_infos:
        context.dataset_info[dataset_info.id] = dataset_info
    for dataset_info in dataset_infos:
        sample_video_dataset(
            api=api,
            src_dataset_id=dataset_info.id,
            dst_parent_info=dst_project_info,
            settings=settings,
            sample_info=sample_info,
            context=context,
            datasets_ids_whitelist=datasets_ids,
            items_progress_cb=items_progress_cb,
            video_progress=video_progress,
        )

    return dst_project_info



# # -*- coding: utf-8 -*-
# """
#
# requests_toolbelt.multipart.decoder
# ===================================
#
# This holds all the implementation details of the MultipartDecoder. https://github.com/requests/toolbelt/pull/222
#
# """
#
#
# import sys
# import email.parser
# from requests_toolbelt.multipart.encoder import encode_with
# from requests.structures import CaseInsensitiveDict
#
#
# def _split_on_find(content, bound):
#     point = content.find(bound)
#     return content[:point], content[point + len(bound):]
#
#
# class ImproperBodyPartContentException(Exception):
#     pass
#
#
# class NonMultipartContentTypeException(Exception):
#     pass
#
#
# def _header_parser(string, encoding):
#     major = sys.version_info[0]
#     if major == 3:
#         string = string.decode(encoding)
#     headers = email.parser.HeaderParser().parsestr(string).items()
#     return (
#         (encode_with(k, encoding), encode_with(v, encoding))
#         for k, v in headers
#     )
#
#
# class BodyPart(object):
#     """
#
#     The ``BodyPart`` object is a ``Response``-like interface to an individual
#     subpart of a multipart response. It is expected that these will
#     generally be created by objects of the ``MultipartDecoder`` class.
#
#     Like ``Response``, there is a ``CaseInsensitiveDict`` object named headers,
#     ``content`` to access bytes, ``text`` to access unicode, and ``encoding``
#     to access the unicode codec.
#
#     """
#
#     def __init__(self, content, encoding):
#         self.encoding = encoding
#         headers = {}
#         # Split into header section (if any) and the content
#         if b'\r\n\r\n' in content:
#             first, self.content = _split_on_find(content, b'\r\n\r\n')
#             if first != b'':
#                 headers = _header_parser(first.lstrip(), encoding)
#         else:
#             raise ImproperBodyPartContentException(
#                 'content does not contain CR-LF-CR-LF'
#             )
#         self.headers = CaseInsensitiveDict(headers)
#
#     @property
#     def text(self):
#         """Content of the ``BodyPart`` in unicode."""
#         return self.content.decode(self.encoding)
#
#
# class MultipartDecoder(object):
#     """
#
#     The ``MultipartDecoder`` object parses the multipart payload of
#     a bytestring into a tuple of ``Response``-like ``BodyPart`` objects.
#
#     The basic usage is::
#
#         import requests
#         from requests_toolbelt import MultipartDecoder
#
#         response = request.get(url)
#         decoder = MultipartDecoder.from_response(response)
#         for part in decoder.parts:
#             print(part.headers['content-type'])
#
#     If the multipart content is not from a response, basic usage is::
#
#         from requests_toolbelt import MultipartDecoder
#
#         decoder = MultipartDecoder(content, content_type)
#         for part in decoder.parts:
#             print(part.headers['content-type'])
#
#     For both these usages, there is an optional ``encoding`` parameter. This is
#     a string, which is the name of the unicode codec to use (default is
#     ``'utf-8'``).
#
#     """
#     def __init__(self, content, content_type, encoding='utf-8'):
#         #: Original Content-Type header
#         self.content_type = content_type
#         #: Response body encoding
#         self.encoding = encoding
#         #: Parsed parts of the multipart response body
#         self.parts = tuple()
#         self.boundary = MultipartDecoder._find_boundary(content_type, encoding)
#         self._parse_body(content)
#
#     @staticmethod
#     def _find_boundary(content_type, encoding):
#         ct_info = tuple(x.strip() for x in content_type.split(';'))
#         mimetype = ct_info[0]
#         if mimetype.split('/')[0].lower() != 'multipart':
#             raise NonMultipartContentTypeException(
#                 "Unexpected mimetype in content-type: '{0}'".format(mimetype)
#             )
#         for item in ct_info[1:]:
#             attr, value = _split_on_find(
#                 item,
#                 '='
#             )
#             if attr.lower() == 'boundary':
#                 boundary = encode_with(value.strip('"'), encoding)
#         return boundary
#
#     @staticmethod
#     def _fix_first_part(part, boundary_marker):
#         bm_len = len(boundary_marker)
#         if boundary_marker == part[:bm_len]:
#             return part[bm_len:]
#         else:
#             return part
#
#     def _parse_body(self, content):
#         boundary = b''.join((b'--', self.boundary))
#
#         def body_part(part):
#             fixed = MultipartDecoder._fix_first_part(part, boundary)
#             return BodyPart(fixed, self.encoding)
#
#         def test_part(part):
#             return (part != b'' and
#                     part != b'\r\n' and
#                     part[:4] != b'--\r\n' and
#                     part != b'--')
#
#         parts = content.split(b''.join((b'\r\n', boundary)))
#         self.parts = tuple(body_part(x) for x in parts if test_part(x))
#
#     @classmethod
#     def from_response(cls, response, encoding='utf-8'):
#         content = response.content
#         content_type = response.headers.get('content-type', None)
#         return cls(content, content_type, encoding)
#
#
# # This is thrown when the object is being reiterated from another place
# class AlreadyIteratedException(Exception):
#     pass
#
#
# # This is thrown when trying to skip to the next part without finishing to
# # stream the previous one
# class PreviousPartNotFinishedException(Exception):
#     pass
#
#
# class StreamPart(object):
#     def __init__(self, headers, encoding, iterator):
#         self.headers = headers
#         self.encoding = encoding
#         self._iterator = iterator
#         self._started = False
#         self._consumed = False
#         self._content = None
#         self._finished = False
#
#     def __iter__(self):
#         if self._started:
#             raise AlreadyIteratedException()
#         self._started = True
#         for typ, data in self._iterator():
#             if typ == 'done' and data is False:
#                 self._finished = True
#                 break
#             elif typ == 'stream':
#                 yield data
#             else:
#                 raise ImproperBodyPartContentException()
#
#     @property
#     def content(self):
#         if self._consumed:
#             return self._content
#         if self._started:
#             raise AlreadyIteratedException()
#         self._content = b''.join(self)
#         self._consumed = True
#         return self._content
#
#     @property
#     def text(self):
#         return self.content.decode(self.encoding)
#
#
# class MultipartStreamDecoder(object):
#     @classmethod
#     def from_response(cls, response, encoding='utf-8', chunk_size=10 * 1024,
#                       header_size_limit=None):
#         def content():
#             return response.raw.read(chunk_size)
#         content_type = response.headers.get('content-type', None)
#         return cls(content, content_type, encoding, header_size_limit)
#
#     def __init__(self, stream_read_func, content_type, encoding='utf-8',
#                  header_size_limit=None):
#         self.content_type = content_type
#         self.encoding = encoding
#         self._stream_read_func = stream_read_func
#         self._header_size_limit = header_size_limit
#         self._boundary = MultipartDecoder._find_boundary(content_type,
#                                                          encoding)
#         self._splitter = StreamSplitter()
#         self._boundary = b''.join((b'--', self._boundary))
#         self._boundary_split = b''.join((b'\r\n', self._boundary))
#         self._state = 0
#         self._found = False
#         self._started = False
#         self._finished = False
#
#     # Call this to drain stream when error occured, or you decide
#     # not to read all data
#     def close(self):
#         if not self._finished:
#             while True:
#                 try:
#                     data = self._stream_read_func()
#                     if not data:
#                         break
#                 # Protection if _stream_read_func is an generator
#                 except StopIteration:
#                     break
#                 finally:
#                     self._finished = True
#
#     # The instance can be used as an context manager for automatic
#     # draining the stream for re-use
#     def __enter__(self):
#         return self
#
#     def __exit__(self, exc_type, exc_value, exc_traceback):
#         self.close()
#
#     def __iter__(self):
#         if self._started:
#             raise AlreadyIteratedException()
#         self._started = True
#         # This is for guarding against iterating before finishing to
#         # iterate on the current part.
#         _current_stream = None
#         for typ, data in self._iterator():
#             if _current_stream and not _current_stream._finished:
#                 raise PreviousPartNotFinishedException()
#             if typ == 'headers':
#                 _current_stream = StreamPart(
#                     data, self.encoding, self._iterator
#                 )
#                 yield _current_stream
#             else:
#                 raise ImproperBodyPartContentException()
#
#     def _iterator(self):
#         while True:
#             data = self._stream_read_func()
#             # This persumes that if data returned empty once it won't return
#             # anything again (EOS)
#             if not self._found and not data:
#                 self._finished = True
#                 break
#             # This part mimics the _fix_first_part logic from above
#             if self._state == 0:
#                 should_be_empty_or_crlf, self._found = self._splitter.stream(
#                     data, self._boundary, True
#                 )
#                 if should_be_empty_or_crlf:
#                     if should_be_empty_or_crlf != b'\r\n':
#                         raise ImproperBodyPartContentException()
#                     self._state = 1
#                     continue
#                 if self._found:
#                     self._state = 1
#                     continue
#             # Parse the headers
#             elif self._state == 1:
#                 headers, self._found = self._splitter.stream(data, b'\r\n\r\n',
#                                                              True)
#                 if headers:
#                     headers = _header_parser(headers.lstrip(), self.encoding)
#                     headers = CaseInsensitiveDict(headers)
#                     self._state = 2
#                     yield 'headers', headers
#                     continue
#                 # No headers found
#                 if self._found:
#                     headers = CaseInsensitiveDict({})
#                     self._state = 2
#                     yield 'headers', headers
#                     continue
#                 # This is to protect against malformed input where a header
#                 # does not exist in a limit for performence reasons
#                 if self._header_size_limit:
#                     if (self._splitter.leftover_length >
#                             self._header_size_limit):
#                         raise ImproperBodyPartContentException()
#             # Stream the part
#             elif self._state == 2:
#                 stream, self._found = self._splitter.stream(
#                     data, self._boundary_split
#                 )
#                 if stream:
#                     yield 'stream', stream
#                 # boundary_split found, end of part
#                 if self._found:
#                     self._state = 1
#                     yield 'done', False
#                     continue
#
#
# class StreamSplitter(object):
#     def __init__(self):
#         self.leftover = b''
#
#     def stream(self, data, split_data, return_only_full=False):
#         self.leftover += data
#         index = self.leftover.find(split_data)
#         if return_only_full:
#             if index > -1:
#                 ret = self.leftover[:index]
#                 self.leftover = self.leftover[index + len(split_data):]
#                 found = True
#             else:
#                 ret = b''
#                 found = False
#         else:
#             if index > -1:
#                 ret = self.leftover[:index]
#                 self.leftover = self.leftover[index + len(split_data):]
#                 found = True
#             elif len(self.leftover) >= len(split_data):
#                 ret = self.leftover[:-len(split_data)]
#                 self.leftover = self.leftover[-len(split_data):]
#                 found = False
#             else:
#                 ret = b''
#                 found = False
#         return ret, found
#
#     @property
#     def leftover_length(self):
#         return len(self.leftover)



# remote storage download api
# MultipartStreamDecoder has bug
#     def _download_batch(self, paths, is_stream=True):
#         def parse_name(headers):
#             name = part.headers[b'Content-Disposition'].decode("utf-8").split('name="')[1].split('"')[0]
#             return name
#         response = self._api.post('remote-storage.bulk.download', paths, stream=is_stream)
#         with MultipartStreamDecoder.from_response(response, chunk_size=1024 * 1024) as decoder:
#             for part in decoder:
#                 #print(part.headers)
#                 remote_path = parse_name(part.headers)
#                 for stream in part:
#                     yield remote_path, stream
#                 # print(part.content) # Read comment below
#
#     def download_paths(self, remote_paths, local_paths):
#         path_map = {remote: local for remote, local in zip(remote_paths, local_paths)}
#         last_file = ""
#         w = None
#         for current_path, resp_part in self._download_batch(remote_paths, is_stream=True):
#             if last_file != current_path:
#                 if w is not None:
#                     w.close()
#                 w = open(path_map[current_path], 'wb')
#                 last_file = current_path
#             w.write(resp_part)
#         if w is not None:
#             w.close()


# coding: utf-8

from supervisely.sly_logger import logger
from supervisely.annotation.annotation import Annotation
from supervisely.metric.common import log_head, log_line, CONFUSION_MATRIX, UNMATCHED_GT, UNMATCHED_PREDICTIONS
from supervisely.metric.matching import filter_labels_by_name, match_labels_by_iou
from supervisely.metric.metric_base import MetricsBase


class ConfusionMatrixMetric(MetricsBase):

    def __init__(self, class_mapping, iou_threshold):
        if len(class_mapping) < 1:
            raise RuntimeError('At least one classes pair should be defined!')
        self._class_mapping = class_mapping.copy()
        self._iou_threshold = iou_threshold
        self._confusion_matrix = {(cls_gt, cls_pred): 0
                                  for cls_gt in class_mapping.keys()
                                  for cls_pred in class_mapping.values()}
        self._unmatched_gt = {cls_gt: 0 for cls_gt in class_mapping.keys()}
        self._unmatched_pred = {cls_pred: 0 for cls_pred in class_mapping.values()}

    def add_pair(self, ann_gt: Annotation, ann_pred: Annotation):
        labels_gt = filter_labels_by_name(ann_gt.labels, self._unmatched_gt)
        labels_pred = filter_labels_by_name(ann_pred.labels, self._unmatched_pred)
        match_result = match_labels_by_iou(labels_1=labels_gt, labels_2=labels_pred, img_size=ann_gt.img_size,
                                           iou_threshold=self._iou_threshold)
        for match in match_result.matches:
            self._confusion_matrix[match.label_1.obj_class.name, match.label_2.obj_class.name] += 1
        for unmatched_gt_label in match_result.unmatched_labels_1:
            self._unmatched_gt[unmatched_gt_label.obj_class.name] += 1
        for unmatched_pred_label in match_result.unmatched_labels_2:
            self._unmatched_pred[unmatched_pred_label.obj_class.name] += 1

    def get_metrics(self):
        return {CONFUSION_MATRIX: self._confusion_matrix.copy(),
                UNMATCHED_GT: self._unmatched_gt.copy(),
                UNMATCHED_PREDICTIONS: self._unmatched_pred.copy()}

    def get_total_metrics(self):
        return self.get_metrics()

    def log_total_metrics(self):
        def exp_one(arg):
            return str(arg).center(20)

        def exp_arg(args):
            return [exp_one(arg) for arg in args]

        log_line()
        log_head(' Result metrics values for {} IoU threshold '.format(self._iou_threshold))
        log_head(' Confusion matrix ')

        sorted_gt_names = sorted(self._class_mapping.keys())
        pred_names = [self._class_mapping[gt_name] for gt_name in sorted_gt_names]
        logger.info(''.join(exp_arg([''] + pred_names + ['False Negatives'])))
        for gt_name in sorted_gt_names:
            logger.info(''.join([exp_one(gt_name)] +
                                exp_arg([self._confusion_matrix[gt_name, pred_name] for pred_name in pred_names]) +
                                [exp_one(self._unmatched_gt[gt_name])]))
            log_line()
        logger.info(''.join([exp_one('False Positives')] +
                            exp_arg([self._unmatched_pred[pred_name] for pred_name in pred_names]) +
                            [exp_one('0')]))
        log_line()



# coding: utf-8

from typing import Dict, List, Optional

from supervisely.api.module_api import ApiField
from supervisely.api.entity_annotation.figure_api import FigureApi, FigureInfo
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudFigureApi(FigureApi):
    """
    :class:`PointcloudFigure<supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure>` for a single point cloud.
    """

    def create(
        self,
        pointcloud_id: int,
        object_id: int,
        geometry_json: Dict,
        geometry_type: str,
        track_id: Optional[int] = None,
    ) -> int:
        """
        Create new PointcloudFigure of given point cloud object in point cloud with given ID.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param object_id: ID of the object to which the PointcloudFigure belongs.
        :type object_id: int
        :param geometry_json: Parameters of geometry for PointcloudFigure.
        :type geometry_json: dict
        :param geometry_type: Type of PointcloudFigure geometry.
        :type geometry_type: str
        :param track_id: int, optional.
        :type track_id: int, optional
        :return: New figure ID
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pcd_id = 19618685
            object_id = 5565921
            geometry_json = {'points': {'exterior': [[500, 500], [1555, 1500]], 'interior': []}}
            geometry_type = 'rectangle'

            figure_id = api.pointcloud.figure.create(pcd_id, object_id, geometry_json, geometry_type) # 643182610
        """

        return super().create(pointcloud_id, object_id, {}, geometry_json, geometry_type, track_id)

    def append_bulk(
        self,
        pointcloud_id: int,
        figures: List[PointcloudFigure],
        key_id_map: KeyIdMap,
    ) -> None:
        """
        Add VideoFigures to given Video by ID.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param figures: List of point cloud figures to append.
        :type figures: List[PointcloudFigure]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 124976
            meta_json = api.project.get_meta(project_id)
            meta = sly.ProjectMeta.from_json(meta_json)
            key_id_map = KeyIdMap()

            pcd_id = 198703212
            ann_info = api.pointcloud.annotation.download(pcd_id)
            ann = sly.PointcloudAnnotation.from_json(ann_info, meta, key_id_map)
            figures = ann.figures[:5]
            api.video.figure.append_bulk(pcd_id, figures, key_id_map)
        """

        keys = []
        figures_json = []
        for figure in figures:
            keys.append(figure.key())
            figures_json.append(figure.to_json(key_id_map))

        self._append_bulk(pointcloud_id, figures_json, keys, key_id_map)

    def append_to_dataset(
        self,
        dataset_id: int,
        figures: List[PointcloudFigure],
        entity_ids: List[int],
        key_id_map: KeyIdMap,
    ) -> None:
        """
        Add pointcloud figures to Dataset annotations.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :param figures: List of point cloud figures.
        :type figures: List[PointcloudFigure]
        :param entity_ids: List of point cloud IDs.
        :type entity_ids: List[int]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :rtype: :class:`NoneType`
        :Usage example:

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d
            from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudObjectCollection
            from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
            from supervisely.video_annotation.key_id_map import KeyIdMap

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()


            project_id = 17231
            dataset_id = 55875
            pointcloud_id = 19373403
            project = api.project.get_info_by_id(project_id)
            dataset = api.dataset.get_info_by_id(dataset_id)

            class_car = sly.ObjClass('car', Cuboid3d)
            classes = sly.ObjClassCollection([class_car])
            project_meta = sly.ProjectMeta(classes)
            updated_meta = api.project.update_meta(project.id, project_meta.to_json())

            key_id_map = KeyIdMap()

            car_object = sly.PointcloudObject(class_car)
            objects_collection = PointcloudObjectCollection([car_object])

            uploaded_objects_ids = api.pointcloud_episode.object.append_to_dataset(
                dataset.id,
                objects_collection,
                key_id_map,
            )

            position, rotation, dimension = Vector3d(-32.4, 33.9, -0.7), Vector3d(0., 0, 0.1), Vector3d(1.8, 3.9, 1.6)
            cuboid = Cuboid3d(position, rotation, dimension)
            figure_1 = PointcloudFigure(car_object, cuboid)

            api.pointcloud_episode.figure.append_to_dataset(
                dataset.id,
                [figure_1],
                [pointcloud_id],
                key_id_map,
            )
        """

        keys = []
        figures_json = []
        for figure, entity_id in zip(figures, entity_ids):
            keys.append(figure.key())
            figure_json = figure.to_json(key_id_map)
            figure_json[ApiField.ENTITY_ID] = entity_id
            figures_json.append(figure_json)

        return self._append_bulk(
            dataset_id, figures_json, keys, key_id_map, field_name=ApiField.DATASET_ID
        )

    def _convert_json_info(self, info: dict, skip_missing=True):
        return super()._convert_json_info(info, skip_missing)

    def download(
        self,
        dataset_id: int,
        pointcloud_ids: List[int] = None,
        skip_geometry: bool = False,
        **kwargs
    ) -> Dict[int, List[FigureInfo]]:
        """
        Method returns a dictionary with pairs of pointcloud ID and list of FigureInfo for the given dataset ID. Can be filtered by pointcloud IDs.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param pointcloud_ids: Specify the list of pointcloud IDs within the given dataset ID. If pointcloud_ids is None, the method returns all possible pairs of images with figures. Note: Consider using `sly.batched()` to ensure that no figures are lost in the response.
        :type pointcloud_ids: List[int], optional
        :param skip_geometry: Skip the download of figure geometry. May be useful for a significant api request speed increase in the large datasets.
        :type skip_geometry: bool

        :return: A dictionary where keys are pointcloud IDs and values are lists of figures.
        :rtype: :class: `Dict[int, List[FigureInfo]]`
        """
        if kwargs.get("image_ids", False) is not False:
            pointcloud_ids = kwargs["image_ids"]  # backward compatibility
        return super().download(dataset_id, pointcloud_ids, skip_geometry)
